/*
 * congestion.c --  RLC congestion control routines.
 *
 * This file is part of
 * 
 * rlc -- Reliable Multicast data Distribution Protocol
 * 
 * (C) 1996, 1997, 1998 Luigi Rizzo and Lorenzo Vicisano
 *     (luigi@iet.unipi.it, vicisano@cs.ucl.ac.uk)
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Luigi Rizzo,
 *      Lorenzo Vicisano and other contributors.
 * 4. Neither the name of the Authors nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include "rlcif.h"
#include "agent.h"
#include "congestion.h"
#include "receiver.h"
#include "layers.h"
#include "utils.h"
#include "error.h"

/* todo:
 * - possibly all the time values can be converted from double to
 *   ui32
 */

/*++++

Enables logging facilities in this receiver.

Params:
	pi : protocol instance descriptor.
	ev_fn : pathname for the event-log file
	po_fn : pathname for the poll-log file

Returns:
	0 on success, < 0 on failure.
*/
int
rlc_rx_log(int pi, char *ev_fn, char *po_fn)
{
  FILE *tev_fd, *tpo_fd;
  cong_t *inst;
    inst = get_cong_inst(pi);
    if (!inst)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    /* check event file name */
    tev_fd = NULL;
    if (ev_fn != NULL) {
	tev_fd = fopen(ev_fn, "w");
	if (!tev_fd)
	    return (-1);
    }

    /* check poll file name */
    tpo_fd = NULL;
    if (po_fn != NULL) {
	tpo_fd = fopen(po_fn, "w");
	if (!tpo_fd)
	    return (-1);
    }

    inst->pollf = tpo_fd;
    inst->eventf = tev_fd;

    return (0);
}


/*++++

Retrieves some status information for this receiver. The information
are in the form defined by the structure  rx_info_t (defined in
rlcif.h). In the current implementation rx_info_t is defined as:

    typedef struct rx_info_s {              
            int nlay;               / number of layers provided by the sender /
            int clay;               / number of layers currently joined /
            unsigned long t0;       / packet intertime extimation on l0 (us) /
            int received;           / total number of packets received /
            int lost;               / total number of packets lost /
            int cong_ev;            / congestion events seen so far /
    } rx_info_t;

Params:
	pi : protocol instance descriptor.
	buf : buffer to store the retrieved
		information.
	len : must be initialized to the buffer
		size and will be set to the size of
		the portion of the buffer used to store
		the retrieved data. If the buffer is
		smaller than the available data size,
		data will be truncated.

Returns:
	0 on success, < 0 on failure.
*/
int
rlc_rx_info(int pi, void *buf, int *len)
{
  rx_info_t str;
  cong_t *inst;
    NDEB(fprintf(stderr, "rlc_rx_info pi = %d\n", pi);)
    inst = get_cong_inst(pi);
    assert(inst != NULL);

    if (buf == NULL || len <= 0)
	return(-1);

    bzero(&str, sizeof(str));

    str.nlay = inst->n_lays;
    str.clay = inst->curr_session_;
    str.t0 = inst->base_period_ * 1e6;
    str.received = inst->tot_count_;
    str.lost = inst->lost_;
    str.cong_ev = inst->cong_ev_;

    *len = min(*len, sizeof(str));
    bcopy(&str, buf, *len);
    /* *len = 0; */
    return(0);
}

int
rlc_rx_set_stats_deliver_f (int pi, int gid, rx_stats_deliver_f stats_deliver,
	int maxsize)
{
  cong_t *inst;
    inst = get_cong_inst(pi);
    if (!inst)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    if (!stats_deliver) {		/* disable */
	inst->stats_deliver = NULL;
	inst->statsbufoffset = 0;
	return 0;
    }

    if (maxsize > 0)
	inst->maxstatbufsize = min (maxsize, MAXSTATSBUFSIZE);
    else
	inst->maxstatbufsize = MAXSTATSBUFSIZE;

    inst->stats_deliver = stats_deliver;
    inst->stats_gid = gid;

    return 0;
}


/*++++
It gets called once at library initialization time.

Params:
	pi : protocol instance descriptor.
	inst : congestion control data structure
		of this protocol instance.

Returns:
	0 on success, < 0 on failure.
*/
int
congestion_setup(int pi, cong_t *inst)
{

    if (!inst)
	return(-1);
    NDEB(fprintf(stderr, "congestion_setup pi = %d\n", pi);)

    inst->local_id_ = pi;
	
    fset(inst->flags, LP_UNS);
    inst->n_lays = UNSEEN;

    inst->thresh_ = inst->init_thresh_ = -1;	/* thresh_ is initially set to
						   max_bwt_/5 */
    fclear(inst->flags, JN_PND);
	
#if UNSUPP
    for (int i=0; i<= inst->start_session_; i++)
	GO_UP(pi);
#endif
    inst->max_bwt_ = 0;
    inst->clean_count_ = inst->tot_count_ = 0;
    inst->lost_ = 0;
    inst->cong_ev_ = 0;
    inst->lost_in_burst = 0;
    fclear(inst->flags, SEEN_UPS);
    fclear(inst->flags, SEEN_BST);
    fclear(inst->flags, LV_PND);
    inst->leave_deaf_ = 0.;

    inst->del_in = inst->del_c = 0;
    inst->del_l = 4;

    inst->lay_limit = UNSEEN;
    inst->pollf = NULL;
    inst->eventf = NULL;

    inst->base_period_ = L0_PERIOD / 1e6;	/* initial estimation */
    inst->last_0_time = 0;
    inst->last_0_idx = 0;

    inst->stats_deliver = NULL;
    inst->statsbufoffset = 0;

    return(0);
}




 /*
 ** should conform to ns format:
 ** event time src dst pktType size flags class src.sport dst.dport seqno uid
 ** The following event are defined here:
 ** +j<l> start joining layer `l'
 ** -j<l> layer `l' joined
 ** +l<l> start leaving layer `l'
 ** -l<l> layer `l' left
 */

char trace_buffer[512];

static void trace_dump(FILE *fil)
{
    if (fil) fprintf(fil, "%s\n", trace_buffer);

}

#if __STDC__
static void mytrace (FILE *fil, char const *fmt, ...)
#else
static void mytrace (fmt, va_alist)
  char *fmt;
  va_dcl
#endif
{
  va_list ap;
    if(fil) {
#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	vsprintf(trace_buffer, fmt, ap);
	va_end(ap);
	trace_dump(fil);
    }
}

/*++++

Sets a limit to the maximum subscription level in this receiver.

Params:
	pi : protocol instance descriptor.
	limit : the maximum subscription level.

Returns:
	0 on success, < 0 on failure.
*/
int
rlc_rx_limit(int pi, int limit)
{
  cong_t *inst;
#ifdef HAVE_UI
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
#endif
    DEB(fprintf(stderr, "rlc_rx_limit pi = %d\n", pi);)
    inst = get_cong_inst(pi);
    assert(inst != NULL);
    if (limit < 0)
	return(-1);

    if (inst->n_lays != UNSEEN && limit >= inst->n_lays) {
	inst->lay_limit = inst->n_lays-1;
    } else {
	if (inst->lay_limit == limit) return 0;
	inst->lay_limit = limit;
    }
#ifdef HAVE_UI
    if (sxs->have_gui)
	rlc_gui_tcl_exec (pi, NULL, "set_limit %d", inst->lay_limit);
#endif
    if (inst->curr_session_ == UNSEEN) return(0);
    while (inst->curr_session_ > inst->lay_limit) {
	do_leave(pi, inst->curr_session_);
	fclear(inst->flags, JN_PND);
	fset(inst->flags, LV_PND);
	inst->curr_session_--;
	fset(inst->flags, LP_UNS);
    }

    return(inst->lay_limit);
}

static void
add_compact_stats(cong_t *inst, ui32 this_time, int force)
{
   int size = sizeof (comp_stats_t);
   comp_stats_t *p = (comp_stats_t *)(inst->statsbuf + inst->statsbufoffset);

    inst->statsbufoffset += size;

    p->time = htonl (this_time);
    p->tot_count = htons ((unsigned short)(inst->tot_count_));
    p->clean_count = htons ((unsigned short)(inst->clean_count_));
    p->lost = htons ((unsigned short)(inst->lost_));
    p->cong_ev = htons ((unsigned short)(inst->cong_ev_));
    p->max_bwt = htons ((unsigned short)(inst->max_bwt_/(1024*8)));
    p->curr_session = (unsigned char)inst->curr_session_;
    p->flags = inst->flags;

    if (force || inst->statsbufoffset + size > inst->maxstatbufsize) {
	inst->stats_deliver(inst->stats_gid,
	    inst->statsbuf, inst->statsbufoffset);
	inst->statsbufoffset = 0;
    }
}
	

/*
 * Leave all groups and possibly compute statistics
 */
void
congestion_end(int pi, cong_t *inst, timest_t endtime)
{
    while (inst->curr_session_ >= 0)
	do_leave(pi, inst->curr_session_--);
    if (inst->eventf) fflush(inst->eventf);
    if (inst->pollf) fflush(inst->pollf);

    if (inst->stats_deliver) {
	add_compact_stats(inst, endtime, 1);
    }
}

#define TH_RESET	-2

#define DUNNO 999999
#define INSEQ 0

static void t0_guess (cong_t *inst, double time, unsigned short idx)
{
    if (idx - inst->last_0_idx == 1 &&
	time - inst->last_0_time < 100. * inst->base_period_ &&
	time - inst->last_0_time > inst->base_period_ / 100. ) {
	inst->base_period_ = inst->base_period_ * 0.8 +
	    (time - inst->last_0_time) * 0.2 ;
    }
    inst->last_0_time = time;
    inst->last_0_idx = idx;
}


static int outseq(ui32 idx, cong_t *inst)
{
#if 1
  int diff = ui32diff(idx, inst->last_p_+(ui32)1);
#else
  int diff = idx - inst->last_p_ -1;
#endif

	if (fisset(inst->flags, LP_UNS)) {
	    fclear(inst->flags, LP_UNS);
	    inst->last_p_ = idx;
	    return DUNNO;
	}

	if (diff < 0)			/* dup or out of sequence!!
					   don't update */
	    return diff;

	inst->last_p_ = idx;
	return diff;
}

void congestion_control(int pi, cong_t *inst, rlc_h *p, int paylen,
	int this_lay, ui32 gidx, timest_t rec_time)
{

	double now;
	int in_burst, burst_starting, burst_ended,
	    burst_ending;
	int inseq, update, minus;
	ui32 this_idx;

	/* compute the global index on all layers, for reordering
	   pourpose */
	this_idx = gidx;

	/* first packet initializations */
	if (inst->n_lays == UNSEEN) {
#ifdef HAVE_UI
		sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
#endif
		now = rec_time/1e6;
		inst->next_poll = now + POLL_INTERVAL;
		assert (p->layer_n <= MAX_LAYERS);
		inst->n_lays = p->layer_n;
#ifdef HAVE_UI
		if (sxs->have_gui)
		    rlc_gui_tcl_exec (pi, NULL, "set_last %d", inst->n_lays-1);
#endif
		NDEB(fprintf(stderr, "--- *** have %d layers\n", inst->n_lays);)
		if (inst->lay_limit == UNSEEN ||
		    inst->lay_limit >= inst->n_lays) {
		    inst->lay_limit = inst->n_lays-1;
#ifdef HAVE_UI
		    if (sxs->have_gui)
			rlc_gui_tcl_exec (pi, NULL, "set_limit %d",
			    inst->n_lays - 1);
#endif
		}
		inst->pkt_size_ = paylen;
#if 0
		mytrace(inst->eventf,
		"v %.6f rlc/br%.2f %d %u %x %d ---- %d %x.%d %x.%d %d %d",
		now, inst->base_rate_/1024., inst->local_id_,
		0, gidx,
		paylen, this_lay, 0, 0, 0, 0,
		this_idx, inst->curr_session_);
#endif
	}

	/*
	**	Packet delay for resequencing
	*/
	inst->heads[inst->del_c] = p;
	inst->rec_times[inst->del_c] = rec_time;
	inst->gidx[inst->del_c] = gidx;
	inst->this_lay[inst->del_c] = this_lay;
	if (inst->del_in < inst->del_l) inst->del_in++;
	NDEB(fprintf(stderr, "%d pakts in buf (%d %d)\n",
	    inst->del_in, inst->del_c, inst->del_l));
	if (inst->del_in < inst->del_l) {
	    inst->del_c++;
	    return;
	} else {
	    int i, j;
	    for (j = i = 0; i < inst->del_l; i++) {
		if (ui32diff(inst->gidx[i], inst->gidx[j]) < 0)
		    j = i;
	    }
	    inst->del_c = j;
	    NDEB(fprintf(stderr, "peeking %d (%d)\n",
		inst->del_c, inst->del_l));
	}
	p = inst->heads[inst->del_c];
	rec_time = inst->rec_times[inst->del_c];
	this_lay = inst->this_lay[inst->del_c];
	gidx = inst->gidx[inst->del_c];

	inst->tot_count_ ++;
	now = rec_time/1e6;

	minus = 0;
	if (fisset(inst->flags, JN_PND) && this_lay != inst->curr_session_)
	    minus = 1;
	/* now compute the index relatively to the current subscription
		level */
	this_idx = global_index (ntohs(p->lsn), ntohs(p->pn),
		this_lay, inst->curr_session_-minus,
		is_in_burst(p->flags));

	NDEB(fprintf(stderr, "--- have packet %lu\n", this_idx);)

	/* `join_pending_' takes into account the delay in joining
	 * sessions, so that we avoid seeing false losses while
	 * waiting for the join to take effect
	 */
	if (fisset(inst->flags, JN_PND) && this_lay == inst->curr_session_) {
							/* join committed */
		    fclear(inst->flags, JN_PND);
		    fset(inst->flags, LP_UNS);
		    mytrace(inst->eventf,
		    "v %.6f rlc/-j%d %d %u %x %d ---- %d %x.%d %x.%d %lu %d",
		    now, inst->curr_session_, inst->local_id_,
		    0, gidx,
		    paylen, this_lay, 0, 0, 0,
		    0, this_idx, inst->curr_session_);
	}

	/* `leave_pending_' takes into account the delay in leaving
	 * sessions: packets coming from the left layer are ignored
	 * and we will prevent cascade leaves
	 */
	if (this_lay > inst->curr_session_) {	/* packet from left layers :
						   ignore ... */
		if (fisset(inst->flags, LV_PND)) {	/* extend timeout */
		    /* don't see why I don't use the same as the first... */
		    int temp = max(0, inst->curr_session_-1);
		    inst->leave_timeout = now + 3. * 
			inst->base_period_ / (1<<temp);
		} else {
		    /* should log a warning: our leave delay evaluation
			doesn't work very well ! */
		}
		free(p);
		return;
	} else if (fisset(inst->flags, LV_PND) && inst->leave_timeout < now) {
						/* leave committed */
		fclear(inst->flags, LV_PND);
		mytrace(inst->eventf,
		"v %.6f rlc/-l%d %d %u %x %d ---- %d %x.%d %x.%d %lu %d",
		now, inst->curr_session_+1, inst->local_id_,
		0, gidx, 
		paylen, this_lay, 0, 0, 0, 0,
		this_idx, inst->curr_session_);
	}

	DEB(
	    mytrace(inst->eventf,
	    "v %.6f rlc/R %d %u %x %d ---- %d %x.%d %x.%d %lu %d - %d %d",
	    now, inst->local_id_,
	    0, gidx, 
	    paylen, this_lay, 0, 0, 0, 0,
	    this_idx, inst->curr_session_, ntohs(p->pn), ntohs(p->lsn));
	)

	/* this extracts pattern information from the (unique)
	** packet sequence # . Extracts a per-session seq. #
	** (this_idx) and the following:
	*/
	in_burst = 0;			/* are we in burst? */
	burst_starting = 0;		/* first burst packet? */
	burst_ended = 0;		/* first non-burst packet? */
	burst_ending = 0;		/* last burst packet? */
	burst_ended = is_first_non_burst ( this_lay,
		inst->curr_session_-minus, p->flags);
	if ((in_burst = is_in_burst(p->flags)) == 1) {
	    burst_starting = is_first_burst (
		this_lay, inst->curr_session_-minus, p->flags);
	    burst_ending = is_burst_ending (p->flags);
	    if (up_synch_burst (this_lay, inst->curr_session_-minus,
		p->flags)) {
		fset(inst->flags, SEEN_UPS);
		NDEB(fprintf(stderr, "=== UP_SYN on lay %d at %.6f (%x)\n",
		    inst->curr_session_, now, gidx);)
	    }
	}
	NDEB(fprintf(stderr, "%d(%d) %d %d %d %d %d - %x - %.6f\n",
		this_lay, inst->curr_session_, burst_starting,
		burst_ending, burst_ended, in_burst,
		fisset(inst->flags, SEEN_UPS), gidx, now);)

	inst->clean_count_ ++;

	/*
	** check the session index to detect packet loss.
	*/
	inseq = outseq(this_idx, inst);
	if (inseq < 0) {
	    /* os = out of seq. (0=dup) */
	    mytrace(inst->eventf,
	    "v %.6f rlc/os%d %d %u %x %d ---- %d %x.%d %x.%d %lu %d",
	    now, inseq+1, inst->local_id_,
	    0, gidx, 
	    paylen, this_lay, 0, 0, 0, 0,
	    this_idx, inst->curr_session_);
	    inseq = INSEQ;
	}

	/* properly adj. in_burst so that:
	**
	** in_burst =   ... 0   0   01  11  11  11                 1   0   0 ...
	** packets      ... X   X   XX  XX  XX  XX                 X   X   X ...
	*/
	if (burst_starting) {
	    /* this packet belong to the `non burst' phase, if
	       we don't see it, we will go down at the next received
	       packet.... */
	    in_burst = 0;
	    inst->lost_in_burst = 0;
	    fset(inst->flags, SEEN_BST);
#if 0
	} else if (burst_ended) {
#else
	} else if (!in_burst) {		/* this seems more correct */
#endif
	    fclear(inst->flags, SEEN_BST);
	}

	if (this_lay == 0 && !in_burst)
	    t0_guess(inst, now, ntohs(p->lsn));


	update = 1;
	if (inseq == DUNNO) {		/* DUNNO, do nothing */
		NDEB(fprintf(stderr, "%d DUNNO [c=%d, l=%d]\n",
			inst->local_id_, inst->curr_session_, this_lay);)
	} else if (inseq == INSEQ) {	/* no nosses */
	    if (!in_burst) {
		/* do nothing ? */
		NDEB(fprintf(stderr, "%d noburst, no loss [c=%d, l=%d]\n",
			inst->local_id_, inst->curr_session_, this_lay);)
	    } else {
		NDEB(fprintf(stderr, "%d burst, no loss [c=%d, l=%d]\n",
			inst->local_id_, inst->curr_session_, this_lay);)

		if (this_idx%2) {		/* just received the second
						   packet of a burst, compute
						   max_bwt_ */
		    int tmp = (int)(inst->pkt_size_ * 8 / 
			(now-inst->last_arrival));
		    if (inst->max_bwt_ < tmp) {
			inst->max_bwt_ = tmp;
			mytrace(inst->eventf,
				"v %.6f rlc/bw%.2f %d %u %x %d ---- %d %x.%d "
				"%x.%d %lu %d", now, inst->max_bwt_/ 1024.,
				inst->local_id_, 0, gidx,
				paylen, this_lay, 0, 0, 0, 0,
				this_idx, inst->curr_session_);

			/* compute thresh_ */
			if (inst->thresh_ != TH_RESET) {
#if 0		/* what to do with this ? */
			    int i;
			    inst->thresh_ = inst->n_lays;
			    for (i=1; i< inst->n_lays; i++)
				if (inst->base_rate_ * (1<<i) >
				    inst->max_bwt_/5) {
				    inst->thresh_ = i-1;
				    break;
				}
			    mytrace(inst->eventf, 
				"v %.6f rlc/th%d %d %u %x %d ---- "
				"%d %x.%d %x.%d %lu %d",
				now, inst->thresh_, inst->local_id_,
				0, gidx,
				paylen, this_lay, 0, 0, 0, 0,
				this_idx, inst->curr_session_);
#endif
			}
		    }
		}
		/* if (last burst and can go up) join next */
		if (burst_ending) {
						/* this is the sync. point,
						   the only one when we are
						   allowed to go up */
		    if ((!inst->lost_in_burst) && (
			(inst->curr_session_ < inst->thresh_) ||
							/* this is only the
							   first rise...
							   slow start ! */
			(fisset(inst->flags, SEEN_UPS) &&
			 !fisset(inst->flags, JN_PND)) )
		       ) {
			fclear(inst->flags, SEEN_UPS);
			if (inst->curr_session_ < inst->lay_limit) { /* GOUP */
			    inst->curr_session_++;
			    if (do_join(pi, inst->curr_session_) < 0) {
				rlc_have_error_ (pi, JOIN_ERROR, NULL);
				return;
			    }
			    fset(inst->flags, JN_PND);
			    fclear(inst->flags, LV_PND);
			    inst->leave_deaf_ = now;
			    mytrace(inst->eventf,
				"v %.6f rlc/+j%d %d %u %x %d ---- "
				"%d %x.%d %x.%d %lu %d",
				now, inst->curr_session_, inst->local_id_,
				0, gidx, paylen, this_lay, 0, 0, 0,
				0, this_idx, inst->curr_session_);
			}
		    }
		}
	    }
	} else {			/* we have had `inseq' losses */

	    inst->lost_ += inseq;

	    mytrace(inst->eventf,
		"v %.6f rlc/ls%d %d %u %x %d ---- %d %x.%d %x.%d %lu %d",
		now, inseq, inst->local_id_, 0, gidx,
		paylen, this_lay, 0, 0, 0, 0,
		this_idx, inst->curr_session_);

	    /* we must be _very_ careful to attribute the losses... */
	    if (in_burst && fisset(inst->flags, SEEN_BST)) {
		inst->lost_in_burst+= inseq;
	    } else {
		/* _NEVER_ _RESTART_ */
		/* XXX we must make sure that a leave, possibly issued
			before, takes effect */
		if (!fisset(inst->flags, LV_PND) && now >= inst->leave_deaf_ &&
		      inst->curr_session_ > 0) {
		    inst->cong_ev_ ++;
		    inst->thresh_ = TH_RESET;
		    fclear(inst->flags, SEEN_UPS);
		    /* GO_DOWN */
		    if (do_leave(pi, inst->curr_session_) < 0) {
			rlc_have_error_ (pi, LEAVE_ERROR, NULL);
			return;
		    }
		    fclear(inst->flags, JN_PND); fset(inst->flags, LV_PND);
		    inst->curr_session_--; fset(inst->flags, LP_UNS);
		    inst->leave_timeout = now + 5. * inst->base_period_ /
			(1<<inst->curr_session_);
		    inst->leave_deaf_ = now + LEAVE_DEAF;
		    mytrace(inst->eventf, "v %.6f rlc/+l%d %d %u %x %d ---- "
			"%d %x.%d %x.%d %lu %d", now, inst->curr_session_+1,
			inst->local_id_, 0, gidx, paylen, this_lay, 0, 0, 0,
			0, this_idx, inst->curr_session_);

		    update = 0;
		}
	    }
	}

	if (inst->next_poll < now) {
	    mytrace(inst->pollf,
		"v %.6f rlcst %d %d %d %d %d %d %d %.4f %d %d %d %d", 
		now, 0, inst->tot_count_, inst->lost_,
		inst->curr_session_, fisset(inst->flags, LV_PND),
		fisset(inst->flags, JN_PND), inst->pkt_size_,
		inst->base_period_, inst->max_bwt_,
		inst->thresh_, inst->clean_count_, inst->cong_ev_);
	    if (inst->stats_deliver) {
		add_compact_stats(inst, rec_time, 0);
	    }
	    inst->next_poll += POLL_INTERVAL;
	}

	free(p);

	if (!update) return;

	inst->last_arrival = now;
}
