/*
 * library header file for rlc
 * 
 * This file is part of
 *
 * rlc -- Reliable Multicast data Distribution Protocol
 * 
 * (C) 1996, 1997, 1998 Luigi Rizzo and Lorenzo Vicisano
 *     (luigi@iet.unipi.it, vicisano@cs.ucl.ac.uk)
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Luigi Rizzo,
 *      Lorenzo Vicisano and other contributors.
 * 4. Neither the name of the Authors nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#ifndef RLCIF_H
#define RLCIF_H

#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#define MAX_RLC_INSTANCES        256

/*
** Error codes, Error infos in error.c.
*/
#define ALLOC_ERROR			-100
#define ALLOC_TIMIT_ERROR		-101
#define BAD_ADDRESS			-102
#define ADDR_NOT_IN_CLASSD		-103
#define BAD_CALLBACK_FUNCTION		-104
#define TOO_MANY_INSTANCES		-105
#define CANT_START_SENDER		-106
#define STOP_TRANSMISSION		-107
#define NO_EVLIST			-108
#define BAD_TTL				-109
#define BAD_NUMBER_OF_LAYS		-110
#define BAD_MIN_PAYLOAD			-111
#define BAD_MAX_PAYLOAD			-112
#define BAD_PERIOD			-113
#define ADDR_IN_USE			-114
#define TOO_MANY_IOS			-115
#define NO_CHAN				-116
#define CANT_START_RECEIVER		-117
#define JOIN_ERROR			-118
#define TOO_MANY_SOCKET			-119
#define NET_CALL_ERROR			-120
#define LEAVE_ERROR			-121
#define TOO_LARGE_PACKET		-122
#define TCL_UI_ERROR			-123
#define BAND_NOT_IMPL			-124
#define TRIG_NOT_IMPL			-125
#define ALLOC_PKT_ERROR			-126
#define ALLOC_TMP_ERROR			-127
#define RECEPTION_TIMEOUT		-128
#define CANT_SET_WATCHDOG		-129
#define NO_SUCH_INSTANCE		-130
#define CANT_CHANGE_ADDR		-131
#define NO_GUI_ERROR			-132

/*
** Warning codes, Warning infos in error.c.
*/
#define	NOUI_WARNING			1000
#define STOP_UI				1001
#define RCV_ERROR			1002
#define SND_ERROR			1003
#define MULTI_BIND_ERROR		1004
#define ODD_PORT			1005
#define NODEFERR_WARNING		1006
#define NOSIGINT_WARNING		1007
#define RECEPTION_TIMEOUT_WAR		1008

#define ERR_BASE			100
#define WARN_BASE			1000
#define ERR_SIZE			33
#define WARN_SIZE			9
#define RLC_IS_ERR(code)	((code) < 0 ? 1:0)
#define IS_RLC_ERR(code)	( (code) <= -100 && (code) > -100 - ERR_SIZE ?\
				 1 : 0 )
#define IS_RLC_WARN(code)	( (code) >= 1000 && (code) < 1000 + WARN_SIZE?\
				 1 : 0 )
#define ERR_MSG_IDX(code)	(RLC_IS_ERR(code) ? (-(code)-100) : ((code)-1000))

/*
** Structs used as function parameters ...
*/

typedef struct rx_info_s {
	int nlay;		/* number of layers provided by the sender */
	int clay;		/* number of layers currently joined */
	unsigned long t0;	/* packet intertime extimation on l0 (us) */
	int received;		/* total number of packets received */
	int lost;		/* total number of packets lost */
	int cong_ev;		/* congestion events seen so far */
	/* anything else to add ? */
} rx_info_t;

typedef struct _addr_list_t {
	int nadd;		/* number od addresses provided */
	unsigned long *addrs;	/* array of addresses */
	short port;		/* port number */
	int ttl;		/* ttl */
} addr_list_t;

typedef struct _rlc_tx_params_t {
	int nlay;		/* number of layers to use */
	int mpkts;		/* minimum packet size ... deprecated */
	int Mpkts;		/* Maximum packet size */
	unsigned long t;	/* interpacket period at the fastest rate */
} rlc_tx_params_t;

#define GET_SESSION_START	-1
#define GET_SESSION_END		-2

/*
** Functions used as function parameters ...
*/

/* this must be implemented by the client */
/*++++
 This function should be implemeted by the client. The function will
 be called by rlc to notify the client of the occurence of an unrecoverable
 error. The client can get the error code and textual info using
 rlc_error_info() . pi is the procol instance descriptor or -1 if a
 general error occurred.
*/
typedef void (*error_handle_t) (int pi, int errcode);

/* watchdog callback ... */
typedef void (*rx_watchdog_t)(int pi);

/* stats deliverer ... */
typedef void (*rx_stats_deliver_f)(int gid, void *buff, int size);

/* sender ... */
/*++++
 This function must be implemeted by the sender client, and a reference
 to it must be passed to rlc at the initialization. The function will
 be called by rlc each time a new packet must be transmitted.

Params:
	pi: client protocol instance ID.
	lay: the layer in which the packet will be transmitted.
	flags: TBD
	p: the payload of the packet being transmitted.

Returns:
 	the payload size, or 0 indicating that the transmission
 	must be terminated, and the current session deleted. A
 	negative value can be returned as well, in which case a
 	dummy packet od size mpkts will be transmitted.
*/
typedef int (*rlc_tx_packet_t)
	(int pi, int lay, char flags, char **p);

/* receiver ... */
/*++++
 This function must be implemeted by the receiver client, and a reference
 to it must be passed to rlc at the initialization. The function will
 be called by rlc each time a new packet has been received.

Params:
	pi: client protocol instance ID.
	lay: the layer in which the packet has been received.
	flags: TBD
	p: the payload of the packet received.
	size: the size of the packet received.
*/
typedef void (*rlc_rx_packet_t)
	(int pi, int lay, char flags, char *p,
	int size);

/*++++
 This function is only used when the client controls directly the
 event loop (without using `rlc_getevent'). It should be implemented by
 the client application, and is called by rlc to add/remove an IO
 descriptor that will be handled by the client.  If the argument `f' is
 non NULL, the descriptor (`iodes') must be added to the client list,
 and the function f must be called when the descriptor is ready for
 reading. If `f' is a null pointer, the descriptor has to be removed
 from the client list.

 All the calls to this function provide a pointer to the same
 function (if f is non NULL), hence a single instance of its value
 can be stored by the client.
*/
typedef void (*rlc_app_chanhandle_t)
	(int iodes, void (f)());

/**
 ** Exported functions prototypes.
 **/

/*
 * Main functions
 */
int rlc_new(int isreceiver);
int rlc_set_addr(int pi, addr_list_t *rlcaddr);
int rlc_attach(int pi, int cpi, void *f);
int rlc_start(int pi);
int rlc_abort(int pi);
int rlc_gui_enable(int pi);
int rlc_gui_geometry(int pi, char *geometry); /* call before enable ! */
int rlc_registerchanhandle(rlc_app_chanhandle_t f);

/* sender specific ... */
int rlc_tx_setup(int pi, rlc_tx_params_t *params);
unsigned long rlc_change_rate(int pi, unsigned long newr);
unsigned long rlc_adj_rate(int pi, double scale);
/* receiver specific ... */
int rlc_rx_info(int pi, void *buf, int *len);
int rlc_rx_limit(int pi, int upper_lay);
int rlc_rx_set_watchdog(int pi, rx_watchdog_t f, int time);
int rlc_rx_log(int pi, char *ev_fn, char *po_fn);
int rlc_rx_set_stats_deliver_f (int pi, int gid,
	rx_stats_deliver_f stats_deliver, int maxsize);
unsigned long rlc_rx_get_src (int pi);

/*
 * UI functions ...
 */
int rlc_gui_loadcallback (int pi, char *name, void *proc,
        void *data, void *delproc);
int rlc_gui_loadtcl (int pi, char *code);
#if __STDC__
int rlc_gui_tcl_exec (int pi, char **ret, char const *fmt, ...);
#else
rlcint _gui_tcl_exec (pi, ret, fmt, va_alist)
  int pi;
  char **ret;
  char *fmt;
  va_dcl;
#endif

/*
 * error handling ...
 */
int rlc_error_info(char **msg);
void rlc_set_error_h(error_handle_t f);
/* these are undocumented !!! */
int rlc_register_error_table (char **table, int base, int length);
int rlc_internal_set_error_ (int pi, int errcode, char const *fmt, ...);
int rlc_return_error_ (int pi, int errcode, char const *fmt, ...);
void rlc_have_error_ (int pi, int errcode, char const *fmt, ...);

/*
 * these are for protocol instance management ...
 */
#define INSTANCE_DES_MEMB		\
    struct _instance_des_t *next;	\
    int ides;				\

typedef struct _instance_des_t {
    INSTANCE_DES_MEMB
} instance_des_t;

int _rlc_insert_instance (instance_des_t *  inst, void *ptr);
void _rlc_return_instance (instance_des_t *  inst, void *ptr);
instance_des_t * _rlc_getinstance(int pi, void *ptr);
int _rlc_init_instance (int maxinstances, void **ptr);
int rlc_get_allinst(int pi, void *ptr);
int _rlc_howmany_instance (void **ptr);

/**
 ** scheduler stuff ....
 **/

typedef int (*CALLBACK)(void *);

#define EV_DEFTIMEOUT   -4      /* got default timeout */
#define EV_GOTSIGNAL    -3      /* got a signal in select() */
#define EV_TERMINATE    -2      /* exit event loop with no timer pending
                                   and no channel to listen to */
#define EV_NO_PENDING   -1      /* exit event loop with no timer pending
                                   but 1 or more channels to listen to
                                   (only if blocking == 0) */
#define EV_NULL         0       /* exit event loop with no event occurred
                                   but some timer pending */

int rlc_getevent(int blocking);
int rlc_insertchan(int chan, int id, CALLBACK f, void *p);
void rlc_deletechan(int chan);
int rlc_settimer(int id, unsigned long delta, CALLBACK f,
	void *p, int real);
void rlc_set_default_timeout(unsigned long delta);
int rlc_showselected ();
int rlc_deletetimer(int id);
int rlc_poolchan(int *rid, void **rp);

/*
 * Misc ...
 */
/* return time in usec */
long timestamp_u();
/* return time in msec */
long timestamp_m();
/* return time in sec */
long timestamp_s();

#endif  /* RLCIF_H */
