/*
 * rlc.c -- sender/receiver setup routines
 *
 * This file is part of
 * 
 * rlc -- Reliable Multicast data Distribution Protocol
 * 
 * (C) 1996, 1997, 1998 Luigi Rizzo and Lorenzo Vicisano
 *     (luigi@iet.unipi.it, vicisano@cs.ucl.ac.uk)
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Luigi Rizzo,
 *      Lorenzo Vicisano and other contributors.
 * 4. Neither the name of the Authors nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>	/* for stat: without this doesn't
			   compile in FreeBSD 2.1 */
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/stat.h>	/* for stat */
#include <unistd.h>	/* for sleep */
#if (defined(Linux))
	#include <sys/uio.h>
#endif
#if (!defined(Linux))
#if (defined(FreeBSD) || (defined(SunOS) && (OSMVER < 5)))
	#include <errno.h>
	#include <sys/uio.h>
#endif
#endif
#include "rlc.h"
#include "rlcif.h"
#include "sender.h"
#include "agent.h"
#include "layers.h"
#include "utils.h"
#include "compat.h"
#include "init.h"
#include "event.h"

static int sender_init = 0;
static ui32 source_id = 0;

#define C(x)	((x!=NULL)?(*x):0)

static void se_delete(sxs_t *s);
static int se_send(sxs_t *s);
static int warm_up(sxs_t *s);

int
rlc_tx_setup(int pi, rlc_tx_params_t *params)
{
  sxs_t *sxs;

    NDEB(fprintf(stderr,
	"rlc_tx_setup: called with %x(%s) %x(%d) %x(%d) %x(%d) %x(%d) %x(%lu)\n",
	(ui32)addr, (addr!=NULL)?addr:"", (ui32)ttl, C(ttl), (ui32)nlay,
	C(nlay), (ui32)mpkts, C(mpkts), (ui32)Mpkts, C(Mpkts), (ui32)t, C(p));)

    sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    if (!sxs)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    /* check nlay */
    if (params->nlay != 0) {
	if (params->nlay < 0 || params->nlay > MAX_LAYERS)
	    return (rlc_return_error_(pi, BAD_NUMBER_OF_LAYS, NULL));
    } else  {
	params->nlay = NLAY;
    }
    NDEB(fprintf(stderr, "rlc_tx_setup: have %d layers\n", tnlay);)

    /* check mpkts */
    if (params->mpkts != 0) {
	if (params->mpkts < MIN_PAYLOAD_SIZE ||
	    params->mpkts > MAX_PAYLOAD_SIZE)
	    return (rlc_return_error_(pi, BAD_MIN_PAYLOAD, NULL));
    } else  {
	params->mpkts = DATA_SIZE;
    }
    NDEB(fprintf(stderr, "rlc_tx_setup: have %d mpkts\n", tmpkts);)

    /* check Mpkts */
    if (params->Mpkts != 0) {
	if (params->Mpkts < MIN_PAYLOAD_SIZE ||
	    params->Mpkts > MAX_PAYLOAD_SIZE ||
	    params->Mpkts < params->mpkts)
	    return (rlc_return_error_(pi, BAD_MAX_PAYLOAD, NULL));
    } else  {
	params->Mpkts = DATA_SIZE;
    }
    NDEB(fprintf(stderr, "rlc_tx_setup: have %d Mpkts\n", tMpkts);)

    /* check t */
    if (params->t != 0) {
	if ((long)(params->t) < 0 || params->t >
	    (double)MAX_L0_PERIOD/(1<<(params->nlay-1)))
	    return (rlc_return_error_(pi, BAD_PERIOD, NULL));
    } else  {
	params->t = (double)MAX_L0_PERIOD/(1<<(params->nlay-1));
    }
    NDEB(fprintf(stderr, "rlc_tx_setup: have %lu t\n", tp);)

    /* add check on throughput and deb. summary */
    NDEB(fprintf(stderr,
	"+++ running at (%.2f, %.2f) pkts/s, (%d, %d) Kbit/s m(m, M), "
	"(%d, %d) Kbit/s M(m, M)\n",
	1.0e6/(params->t*(1<<(params->nlay-1))), 1.0e6/params->t,
	(int)(8 * params->mpkts * 1.0e6 /
	(params->t * 1024 * (1<<(params->nlay-1)))),
	(int)(8 * params->mpkts * 1.0e6 /
	(params->t * 1024)),
	(int)(8 * params->Mpkts * 1.0e6 /
	(params->t * 1024 * (1<<(params->nlay-1)))),
	(int)(8 * params->Mpkts * 1.0e6 /
	(params->t * 1024)));)

    sxs->rate = (double)params->t*(1<<(params->nlay-1));
    sxs->mpkts = params->mpkts;
    sxs->Mpkts = params->Mpkts;
    sxs->l.nc = params->nlay;

    return(0);
}

/*++++

 Adjusts the transmission rate at run time. The change will take
 effect from the next packet transmission. The same constraint for the
 interpacket time holds, as in rlc_tx_setup().

 Warning: this facility must be used carefully, as the congestion control
 algorithm doesn't take into account the intertime modification.

 Params:
	pi : protocol instance ID.
	newr : the new period (in usec) between 2 consecutive packet
		transmissions at the higher rate; see ``tx_setup''
		`p' parameter for more details.

 Returns:
	the new value on success, 0 on failure.
*/
unsigned long
rlc_change_rate(int pi, unsigned long newr)
{
  sxs_t *se = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
  ui32 lr;

    if (!se) return(0);

    DEB(fprintf(stderr, "ooo+++ rlc_change_rate with new = %lu\n", newr);)
    DEB(fprintf(stderr,
	"ooo+++ was running at (%.2f, %.2f) pkts/s, (%d, %d) Kbit/s m(m, M),"
	" (%d, %d) Kbit/s M(m, M)\n",
	1.0e6/se->rate, 1.0e6*(1<<(se->l.nc-1))/se->rate,
	(int)(8 * se->mpkts * 1.0e6 / (se->rate * 1024)),
	(int)(8 * se->mpkts * 1.0e6 * (1<<(se->l.nc-1)) / (se->rate * 1024)),
	(int)(8 * se->Mpkts * 1.0e6 / (se->rate * 1024)),
	(int)(8 * se->Mpkts * 1.0e6 * (1<<(se->l.nc-1)) / (se->rate * 1024)));)
    lr  = newr*(1<<(se->l.nc-1));

    if (lr > MAX_L0_PERIOD) {
	lr = MAX_L0_PERIOD;
	newr = (double)lr/(1<<(se->l.nc-1));
    }
    se->rate = lr;

    l_change_rate(newr, &se->l);
    DEB(fprintf(stderr,
	"ooo+++ now running at (%.2f, %.2f) pkts/s, (%d, %d) Kbit/s m(m, M),"
	" (%d, %d) Kbit/s M(m, M)\n",
	1.0e6/se->rate, 1.0e6*(1<<(se->l.nc-1))/se->rate,
	(int)(8 * se->mpkts * 1.0e6 / (se->rate * 1024)),
	(int)(8 * se->mpkts * 1.0e6 * (1<<(se->l.nc-1)) / (se->rate * 1024)),
	(int)(8 * se->Mpkts * 1.0e6 / (se->rate * 1024)),
	(int)(8 * se->Mpkts * 1.0e6 * (1<<(se->l.nc-1)) / (se->rate * 1024)));)

    return(newr);
}

/*++++

 Adjusts the transmission rate by a scale factor. The change will take
 effect from the next packet transmission. The same constraint for the
 interpacket time holds, as in rlc_tx_setup().

 Warning: this facility must be used carefully, as the congestion control
 algorithm doesn't take into account the intertime modification.

 Params:
	pi : protocol instance ID.
	scale : the scale factor.

 Returns:
	the new value on success, 0 on failure.
*/
unsigned long
rlc_adj_rate(int pi, double scale)
{
  unsigned long newr;
  sxs_t *se = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);

    if (!se)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    if (se->rate / scale <= MAX_L0_PERIOD)
	se->rate = se->rate / scale;
    else
	se->rate = MAX_L0_PERIOD;

    newr = (ui32)((double)se->rate/(1<<(se->l.nc-1)));
    l_change_rate(newr, &se->l);

    NDEB(fprintf(stderr,
	"+++ running at (%.2f, %.2f) pkts/s, (%d, %d) Kbit/s m(m, M), (%d, %d) Kbit/s M(m, M)\n",
	1.0e6/se->rate, 1.0e6*(1<<(se->l.nc-1))/se->rate,
	(int)(8 * se->mpkts * 1.0e6 / (se->rate * 1024)),
	(int)(8 * se->mpkts * 1.0e6 * (1<<(se->l.nc-1)) / (se->rate * 1024)),
	(int)(8 * se->Mpkts * 1.0e6 / (se->rate * 1024)),
	(int)(8 * se->Mpkts * 1.0e6 * (1<<(se->l.nc-1)) / (se->rate * 1024)));)
    return (newr);
}

/*++++

 Starts the transmission. It first transmits 3 warm-up packets on
 all the multicast channels to allow the pruning of the multicast
 tree; then it schedules the first data packet transmission. The
 client-provided function `tx_packet' is called each time a packet
 must be transmitted. The transmissions ends when tx_packet returns
 a zero value.

 Params:
	pi : protocol instance ID.

 Returns:
	0 on success, < 0 on failure, > 0 if a warning
	is generated, use rlc_error_info() to get error
	or warning description. Possible error/warning
	causes are:
		- CANT_START_SENDER (cannot start this sender).
*/
int
_rlc_tx_start(int pi)
{
  int ret;
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
  addr_list_t addrlist;

    if (!sxs)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    /* addrss has been initialized by attach */
    addrlist.nadd = sxs->l.nc;
    addrlist.addrs = (unsigned long *)sxs->addrss;

    if ((ret = setup_l_sock(&sxs->l, &addrlist, sxs->port, sxs->ttl)) < 0) {
	return (rlc_return_error_ (pi, NET_CALL_ERROR, NULL));
    }
#ifdef ONE_SEND_SOCK
    sxs->ch = sxs->l.chan;
#else
    sxs->ch = sxs->l.chs[0];
#endif
    NDEB(fprintf(stderr, "rlc_tx_setup: setup_l_sock done\n"));
   
    setup_layers((ui32)((double)sxs->rate/(1<<(sxs->l.nc-1))),
		&sxs->l, THE_W, THE_Y, THE_P);

    sxs->isreceiver = 0;

    warm_up(sxs);

    sxs->pkts_left = INFINITE;

    NDEB(fprintf(stderr, "rlc_tx_start: pkts_left %d (%d * %d pkt/s)\n",
	sxs->pkts_left, duration_glob,
	(int)(1e6 * (1<<(sxs->l.nc-1)) / sxs->rate));)
    if (rlc_settimer(TX_RUN, sxs->rate, (CALLBACK)se_send,
	(void *)sxs, 1) < 0)
	return (rlc_return_error_(pi, CANT_START_SENDER, NULL));
    return(0);
}

void
_rlc_tx_init()
{
    /* sender specific initialization ? */
    if (!sender_init) { 
	source_id = compute_id();
	sender_init = 1;
    }
}

static char dummyp[MAX_PKT_LEN];

int
rlc_tx_send_appl_pkt (int pi, char *buff, int paylen)
{
    rlc_h *packet = (rlc_h *)dummyp;
    struct msghdr msg;
    struct iovec iov[2];

    sxs_t *s = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);

    if (!s)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    if (!(s->is_on)) return -1;

    if (paylen > s->Mpkts || paylen < 0) {
	return (rlc_return_error_(pi, TOO_LARGE_PACKET, 
	    "attmp to transmit a too large packet (%d)", paylen));
    }

    /* only fill transport header */
    FILL_D_H(*packet, htons(0), htons(0), htonl(source_id),
	0, (ui8)s->l.nc, BT_EXP_1, 1);
    packet->type = RLC_APPL_T;

    bzero((char *)&msg, sizeof(msg));
    msg.msg_iovlen = 2;
    msg.msg_iov = iov;
    iov[0].iov_base = (caddr_t)packet;
    iov[0].iov_len = DATA_H_SIZE;
    iov[1].iov_base = (caddr_t)buff;
    iov[1].iov_len = paylen;
    NDEB(fprintf(stderr,
	"+=+= now %d (%x) next in %lu\n", paylen, buff, next_delta);)
#ifdef ONE_SEND_SOCK
    msg.msg_name = (caddr_t)&(s->l.saddrs[0]);
    msg.msg_namelen = sizeof(s->l.saddrs[0]);
    if (sendmsg(s->l.chan, &msg, 0) == -1)
#else
    if (sendmsg(s->l.chs[0], &msg, 0) == -1)
#endif
	return (rlc_return_error_(pi, SND_ERROR, NULL));
    return (0);
}

/*
** se_delete gets called 2 times: the 1st (asynch) mark the session deleted,
** the second (synch) actually delete the session.
*/
static void
se_delete(sxs_t *s)
{
  int pi = s->ides;
  sxs_t *ss = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    if (!ss) return;
    if (ss != s) return;

    NDEB(fprintf(stderr, "--- se_delete: sess (%x) is_on = %d\n",
	ss->ides, ss->is_on);)
    if (ss->is_on) {
	ss->is_on = 0;
	return;
    }

#ifdef ONE_SEND_SOCK
    close(ss->l.chan);
#else
    { int i;
    for (i=0; i<(ss->l.nc); i++)
	close(ss->l.chs[i]);
    }
#endif

    _rlc_return_instance((instance_des_t *) ss, rlc_inst_mngr);
    free(ss);
    NDEB(fprintf(stderr, "--- se_delete: sess (%x) freed\n", s->ides);)

    return;
}

#define WARM_UP_PKTS 3
static int
warm_up(sxs_t *s)
{
    int i, l;
    ui8 this_flags = 0;
    rlc_h *dummyh = (rlc_h *)dummyp;

    NDEB(fprintf(stderr, "starting warm-up\n"));

    for (i=0; i<WARM_UP_PKTS; i++) {
	for (l=0; l<s->l.nc; l++) {
	    FILL_D_H(*dummyh, 0, 0, htonl(source_id),
		this_flags, (ui8)s->l.nc, BT_EXP_1, 0);
	    dummyh->type = RLC_WRMP_T;
#ifdef ONE_SEND_SOCK
	    if (sendto(s->l.chan, dummyp,
		DATA_H_SIZE, 0,
		(struct sockaddr *)&(s->l.saddrs[l]),
		sizeof(s->l.saddrs[l])) == -1) {
		rlc_have_error_ (s->ides, SND_ERROR, NULL);
	    }
#else
	    if (send(s->l.chs[l], dummyp,
		DATA_H_SIZE, 0) == -1) {
		rlc_have_error_ (s->ides, SND_ERROR, NULL);
	    }
#endif
	}
	sleep(1);
    }
    NDEB(fprintf(stderr, "warm-up finished\n"));
    return 0;
}

static int
se_send(sxs_t *s)
{
    rlc_h *packet = (rlc_h *)dummyp;
    char *buff;
    int paylen, givenlen;
    static int do_err = 1;
    int this_layer;
    ui16 this_pn;
    ui16 this_lidx;
    ui32 next_delta;
    ui8 this_flags = '\0';
    struct msghdr msg;
    struct iovec iov[2];

    if (!(s->is_on)) {
	s->tx_packet(s->cpi, -1, 0, &buff);	/* notify the termination */
	se_delete(s);
	return 0;
    }

    /* only consec. addr. supported for the time being ! */
    if (s->flags & ADDR_CONS)
	this_flags |= RLC_F_CONS_ADDR;
    else
	this_flags &= (~RLC_F_CONS_ADDR);

    /* in this implementation RLC_F_EDGE_TRIG is stucked at 0 */

    next_delta = 
	send_layers(s, &buff, &givenlen, &this_layer, &this_pn, &this_lidx,
		&this_flags);
    if (next_delta == IS_OVER) {
	s->is_on = 0;
	se_delete(s);
	return 0;
    }

    if (givenlen > s->Mpkts) {
	rlc_have_error_(s->ides, TOO_LARGE_PACKET, 
	    "attmp to transmit a too large packet (%d)", givenlen);
	this_flags |= RLC_F_HART_BEAT;
	paylen = s->mpkts;
	buff = dummyp+DATA_H_SIZE;
    } else if (givenlen < 0) {
	this_flags |= RLC_F_HART_BEAT;
	paylen = s->mpkts;
	buff = dummyp+DATA_H_SIZE;
    } else {
	paylen = givenlen;
    }

    NDEB(fprintf(stderr, "--- sending %d bytes (%d) |%d|-[%d -- %d] %x\n",
	paylen, givenlen, this_layer, this_pn, this_lidx, this_flags));

    /* only fill transport header */
    FILL_D_H(*packet, htons(this_lidx), htons(this_pn), htonl(source_id),
	this_flags, (ui8)s->l.nc, BT_EXP_1, 1);

    bzero((char *)&msg, sizeof(msg));
    msg.msg_iovlen = 2;
    msg.msg_iov = iov;
    iov[0].iov_base = (caddr_t)packet;
    iov[0].iov_len = DATA_H_SIZE;
    iov[1].iov_base = (caddr_t)buff;
    iov[1].iov_len = paylen;
    NDEB(fprintf(stderr,
	"+=+= now %d (%x) next in %lu\n", paylen, buff, next_delta);)
#ifdef ONE_SEND_SOCK
    msg.msg_name = (caddr_t)&(s->l.saddrs[this_layer]);
    msg.msg_namelen = sizeof(s->l.saddrs[this_layer]);
    if (sendmsg(s->l.chan, &msg, 0) == -1) {
#else
    if (sendmsg(s->l.chs[this_layer], &msg, 0) == -1) {
#endif
	if (do_err) {
	    rlc_have_error_ (s->ides, SND_ERROR, NULL);
	}
	do_err = 0;
#if 0
	perror ("send");
	/* XXX */
#endif
    } else
	do_err = 1;
    NDEB(fprintf(stderr, "*** data send (%d)\n", do_err));

    if ( (s->pkts_left % 100 == 0)) {
	NDEB(fprintf(stderr, "--- %4d packets left for session 0x%08x\n",
		s->pkts_left, s->ides) );
    }
    if (s->pkts_left != INFINITE)
	s->pkts_left -- ;
    if (s->pkts_left) {
	if (rlc_settimer(s->ides, next_delta,
	    (CALLBACK)se_send, s, 0) < 0) {
	    se_delete(s);
	    rlc_have_error_(s->ides, STOP_TRANSMISSION, NULL);
	}
	NDEB(fprintf(stderr, "--- next d %d\n", next_delta));
    } else {
	s->is_on = 0;
	se_delete(s);
    }
    return 0;
}
